/* Author: Leon Ehrhardt | Universitaet Bremen <s_8vkuts@uni-bremen.de> */

import java.util.function.Function;
import java.util.function.Predicate;

class List {
  public Object elem;
  public List next;

  // current type of list
  private Class<?> classType;

  public List(Object el, List tl) {
    
    // Typecheck
    /********/
    classType = el.getClass();
    if(tl != null) typeCheck(tl.getType()); 
    /********/

    this.elem= el;
    this.next= tl;
  }

  public String toString() {
    List l= this;
    String comma= "";
    String str= "[";

    while (l != null) {
      str= str+ comma+ l.elem.toString();
      comma= ", ";
      l= l.next;
    }
    return (str+ "]");
  }

  int length() {
    int i= 0;
    for (List cur= this; cur != null; cur= cur.next)
      i++;
    return i;
  }

  void concat(List o) 
  {
    /********/
    typeCheck(o.getType());
    /********/

    List cur= this; 
    while (cur.next != null) {
      cur= cur.next;
      }
    cur.next= o;    
  }

/* 
  Higher-order functions use the functional interface of Java:

  https://docs.oracle.com/en/java/javase/17/docs/api/java.base/java/util/function/Function.html
  https://docs.oracle.com/en/java/javase/17/docs/api/java.base/java/util/function/Predicate.html
*/

  // Problem: Filter entfernt eventuell @this@
  <T> List filter(Predicate<T> f, Class<T> type)
  {
    // Typcheck möglich?
    /********/
    // Class<?> tType = ((T)this.elem).getClass(); // Does not work?
    // Class<?> tType2 = ((T)new Object()).getClass(); Does not work?
    typeCheck(type);
    /********/

    List prev= null;
    List ret= null;
    List l= this;

    while (l != null) {
      if (f.test((T)l.elem)) {
        // Keep l, so prev is l now
        prev= l;
        if (ret == null) {
          // l is the first element to keep
          ret= l;
        }
      } else {
        // Remove  l
        if(prev != null) {
          prev.next= l.next;  
      } } 
      l= l.next;
      }
    return ret;              
  }

 // Problem: map kann eigentlich den Typen ändern, also map<T,S>.
  <T,S> void map(Function<T,S> f, Class<T> type)
  {
    /********/
    // Cannot get Class<?> of T because of Typeerasure
    // Class<?> tType = ((T)this.elem).getClass(); // Does is work? Tautology
    // Class<?> tType2 = ((T)new Object()).getClass(); // Does is work? Always objectclass
    typeCheck(type);
    /********/
     
    List cur= this;

    do {
      /********/
      cur.elem= (S)f.apply((T)cur.elem);
      cur.classType =  ((S)cur.elem).getClass();  
      cur = cur.next;
      /********/
      } while(cur != null);
  }

  public Class<?> getType() 
  {
      return classType;
  }

  private void typeCheck(Class<?> type) 
  {
      if(!this.classType.equals(type)) throw new RuntimeException("Typfehler");
  }
}

class ListMain {

  public static void main(String[] args)
  {
    List l1= new List(1, new List(2, null));

    List l2 = new List(4, new List(5, null));

    List l3 = new List('a', null);

    l1.concat(l2);  

    try 
    {
      l1.concat(l3); // Does not work - type error
    } 
    catch(Exception e) 
    {
      System.out.println(e);
    }

    System.out.println("length: "+ l1.length());
    System.out.println("type: "+ l1.getType());
    System.out.println(l1);

    //change type of list
    l1.map(
      (Integer i) -> {
        Character c; 
        switch(i) {
          case 1:
            c = 'a';
            break; 
          case 2:
            c = 'b';
            break; 
          case 3:
            c = 'c';
            break; 
          case 4:
            c = 'd';
            break;
          default: 
            c = 'z';
        }
        return c;
      }, 
      Integer.class
    );

    System.out.println("length: "+ l1.length());
    System.out.println("type: "+ l1.getType());
    System.out.println(l1);

    l1.concat(l3);

    System.out.println("length: "+ l1.length());
    System.out.println("type: "+ l1.getType());
    System.out.println(l1);
  }
}


