
import java.util.function.Function;
import java.util.function.BiFunction;
import java.util.function.Predicate;

class List<T> {
  public T elem;
  public List<T> next;

  public List(T el, List<T> tl) {
    this.elem= el;
    this.next= tl;
  }

  public String toString() {
    List<T> l= this;
    String comma= "";
    String str= "[";

    while (l != null) {
      str= str+ comma+ l.elem.toString();
      comma= ", ";
      l= l.next;
    }
    return (str+ "]");
  }

  int length() {
    int i= 0;
    for (List<T> cur= this; cur != null; cur= cur.next)
      i++;
    return i;
  }

  void concat(List<T> o) 
  {
    List<T> cur= this; 
    while (cur.next != null) {
      cur= cur.next;
      }
    cur.next= o;    
  }

/* 
  Higher-order functions use the functional interface of Java:

  https://docs.oracle.com/en/java/javase/17/docs/api/java.base/java/util/function/Function.html
  https://docs.oracle.com/en/java/javase/17/docs/api/java.base/java/util/function/Predicate.html
*/

  // Problem: Filter entfernt eventuell @this@
  List<T> filter(Predicate<T> f)
  {
    List<T> prev= null;
    List<T> ret= null;
    List<T> l= this;

    while (l != null) {
      if (f.test(l.elem)) {
        // Keep l, so prev is l now
        prev= l;
        if (ret == null) {
          // l is the first element to keep
          ret= l;
        }
      } else {
        // Remove  l
        if(prev != null) {
          prev.next= l.next;  
      } } 
      l= l.next;
      }
    return ret;              
  }

 // Problem: map kann eigentlich den Typen ändern, also map<T,S>.
 <S> List<S> map(Function<T,S> f)
 {
  List<T> cur= this;
  List<S> prev= null;
  List<S> ret= null;

  while (cur.next != null) {
    List<S> el = new List<>(f.apply(cur.elem), null);
    if (prev != null) {
      prev.next= el;
    } else { 
      ret = el; 
    }
    prev= el;
    cur= cur.next; 
  }
  return ret;
 }

<S> S foldl(BiFunction<S, T, S> f, S e) 
{
  List<T> cur= this;

  while (cur != null) {
    e= f.apply(e, cur.elem);
    cur= cur.next;
  }
  return e;
}

// reverse-in-place
List<T> rev() {
  List<T> n;
  List<T> prev= null;
  List<T> cur= this;

  while (cur != null) {
    n = cur.next;
    cur.next= prev;
    prev= cur;
    cur= n;
  }
  return prev;
}

<S> List<S> map1(Function<T,S> f)
{
  List<S> r= foldl((acc, hd) -> new List<S>(f.apply(hd), acc), null);
  return r.rev();
}

/*
<S> List<S> map2(Function<T,S> f)
 {
  class Pair<A,B> {A fst; B snd; Pair(A a, B b) {this.fst= a; this.snd= b; }}; // for crying out loud

  return foldl((ea, hd) -> {List<S> nu= new List<S>(f.apply(hd), null);
                            if (ea.snd != null) { ea.snd.next= nu; }
                            new Pair<>(ea.fst == null ? nu : ea.fst, nu);
                           }, new Pair<>(null, null)).fst;
 }
*/
}

class ListMain {

public static void main(String[] args)
{
List<Integer> l1= new List<>(1, new List<>(2, null));

List<Integer> l2 = new List<>(4, new List<>(5, null));

List<Character> l3 = new List<>('a', null);

l1.concat(l2);  

// l1.concat(l3); // Does not work - type error

System.out.println(l1);

List<Integer> l4 = l1.map1((x)-> x+2);
System.out.println(l4);

l4= l4.filter((x)-> x % 2 == 0);
System.out.println(l4);

List<Integer> l5= l1.rev();
System.out.println(l1);
System.out.println(l5);

}
	    
}
